"""
Usage documentation at: <https://py-pdf.github.io/fpdf2/Tables.html>
"""

from dataclasses import dataclass, replace
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from .drawing_primitives import DeviceCMYK, DeviceGray, DeviceRGB
from .enums import (
    Align,
    CellBordersLayout,
    MethodReturnValue,
    TableBordersLayout,
    TableCellFillMode,
    TableHeadingsDisplay,
    TableSpan,
    VAlign,
    WrapMode,
)
from .errors import FPDFException
from .fonts import CORE_FONTS, FontFace
from .util import ImageType, Number, NumberClass, Padding

if TYPE_CHECKING:
    from .fpdf import FPDF

DEFAULT_HEADINGS_STYLE = FontFace(emphasis="BOLD")


class Table:
    """
    Object that `fpdf.fpdf.FPDF.table()` yields, used to build a table in the document.
    Detailed usage documentation: https://py-pdf.github.io/fpdf2/Tables.html
    """

    def __init__(
        self,
        fpdf: "FPDF",
        rows: Sequence[str] = (),
        *,
        align: str | Align = "CENTER",
        v_align: str | VAlign = "MIDDLE",
        borders_layout: str | TableBordersLayout = TableBordersLayout.ALL,
        cell_fill_color: Optional[
            float | Sequence[float] | str | DeviceRGB | DeviceGray | DeviceCMYK
        ] = None,
        cell_fill_mode: str | TableCellFillMode = TableCellFillMode.NONE,
        col_widths: Optional[float | Sequence[float]] = None,
        first_row_as_headings: bool = True,
        gutter_height: float = 0,
        gutter_width: float = 0,
        headings_style: FontFace = DEFAULT_HEADINGS_STYLE,
        line_height: Optional[Number] = None,
        markdown: bool = False,
        text_align: str | Align = "JUSTIFY",
        width: Optional[Number] = None,
        wrapmode: WrapMode = WrapMode.WORD,
        padding: Optional[Number | Sequence[Number] | Padding] = None,
        outer_border_width: Optional[Number] = None,
        num_heading_rows: Number = 1,
        repeat_headings: TableHeadingsDisplay | str | int = 1,
        min_row_height: Optional[Number] = None,
    ):
        """
        Args:
            fpdf (fpdf.FPDF): FPDF current instance
            rows: optional. Sequence of rows (iterable) of str to initiate the table cells with text content
            align (str, fpdf.enums.Align): optional, default to CENTER. Sets the table horizontal position relative to the page,
                when it's not using the full page width
            borders_layout (str, fpdf.enums.TableBordersLayout): optional, default to ALL. Control what cell borders are drawn
            cell_fill_color (float, tuple, fpdf.drawing.DeviceGray, fpdf.drawing.DeviceRGB): optional.
                Defines the cells background color
            cell_fill_mode (str, fpdf.enums.TableCellFillMode): optional. Defines which cells are filled with color in the background
            col_widths (float, tuple): optional. Sets column width. Can be a single number or a sequence of numbers.
                 When `col_widths` is a single number, it is interpreted as a fixed column width in document units.
                 When `col_widths` is provided as an array, the values are considered to be fractions of the full effective page width,
                 meaning that `col_widths=(1, 1, 2)` is strictly equivalent to `col_widths=(25, 25, 50)`.
            first_row_as_headings (bool): optional, default to True. If False, the first row of the table
                is not styled differently from the others
            gutter_height (float): optional vertical space between rows
            gutter_width (float): optional horizontal space between columns
            headings_style (fpdf.fonts.FontFace): optional, default to bold.
                Defines the visual style of the top headings row: size, color, emphasis...
            line_height (number): optional. Defines how much vertical space a line of text will occupy
            markdown (bool): optional, default to False. Enable markdown interpretation of cells textual content
            text_align (str, fpdf.enums.Align, tuple): optional, default to JUSTIFY. Control text alignment inside cells.
            v_align (str, fpdf.enums.VAlign): optional, default to CENTER. Control vertical alignment of cells content
            width (number): optional. Sets the table width
            wrapmode (fpdf.enums.WrapMode): "WORD" for word based line wrapping (default),
                "CHAR" for character based line wrapping.
            padding (number, tuple, Padding): optional. Sets the cell padding. Can be a single number or a sequence of numbers, default:0
                If padding for left and right ends up being non-zero then c_margin is ignored.
            outer_border_width (number): optional. Sets the width of the outer borders of the table.
                Only relevant when borders_layout is ALL or NO_HORIZONTAL_LINES. Otherwise, the border widths are controlled by FPDF.set_line_width()
            num_heading_rows (number): optional. Sets the number of heading rows, default value is 1. If this value is not 1,
                first_row_as_headings needs to be True if num_heading_rows>1 and False if num_heading_rows=0. For backwards compatibility,
                first_row_as_headings is used in case num_heading_rows is 1.
            repeat_headings (fpdf.enums.TableHeadingsDisplay): optional, indicates whether to print table headings on every page, default to 1.
        """
        self._fpdf = fpdf
        self._table_align = Align.coerce(align)
        self._v_align = VAlign.coerce(v_align)
        self._borders_layout = TableBordersLayout.coerce(borders_layout)
        self._outer_border_width = outer_border_width
        self._cell_fill_color = cell_fill_color
        self._cell_fill_mode = TableCellFillMode.coerce(cell_fill_mode)
        self._col_widths: Optional[float | Sequence[float]] = col_widths
        self._first_row_as_headings: bool = first_row_as_headings
        self._gutter_height = gutter_height
        self._gutter_width = gutter_width
        self._headings_style = headings_style
        self._line_height: float = (
            2 * fpdf.font_size if line_height is None else float(line_height)
        )
        self._markdown = markdown
        self._text_align = text_align
        self._width = width
        self._wrapmode = wrapmode
        self._num_heading_rows = int(num_heading_rows)
        self._repeat_headings = TableHeadingsDisplay.coerce(repeat_headings)
        self._min_row_height = min_row_height
        self._initial_style: Optional[FontFace] = None
        self.rows: List[Row] = []

        if padding is None:
            self._padding = Padding.new(0)
        else:
            self._padding = Padding.new(padding)

        # check table_border_layout and outer_border_width
        if self._borders_layout not in (
            TableBordersLayout.ALL,
            TableBordersLayout.NO_HORIZONTAL_LINES,
        ):
            if outer_border_width is not None:
                raise ValueError(
                    "outer_border_width is only allowed when borders_layout is ALL or NO_HORIZONTAL_LINES"
                )
            self._outer_border_width = None
        # _outer_border_margin : vertical spacing around table
        if self._outer_border_width:
            assert outer_border_width is not None
            self._outer_border_margin = (
                (gutter_width + float(outer_border_width) / 2),
                (gutter_height + float(outer_border_width) / 2),
            )
        else:
            self._outer_border_margin = (0, 0)

        # check first_row_as_headings for non-default case num_heading_rows != 1
        if self._num_heading_rows != 1:
            if self._num_heading_rows == 0 and self._first_row_as_headings:
                raise ValueError(
                    "first_row_as_headings needs to be False if num_heading_rows == 0"
                )
            if self._num_heading_rows > 1 and not self._first_row_as_headings:
                raise ValueError(
                    "first_row_as_headings needs to be True if num_heading_rows > 0"
                )
        # for backwards compatibility, we respect the value of first_row_as_headings when num_heading_rows==1
        else:
            if not self._first_row_as_headings:
                self._num_heading_rows = 0

        for row in rows:
            self.row(row)

    def row(
        self,
        cells: Sequence[str] = (),
        style: Optional[FontFace] = None,
        v_align: Optional[str | VAlign] = None,
        min_height: Optional[float] = None,
    ) -> "Row":
        "Adds a row to the table. Returns a `Row` object."
        if self._initial_style is None:
            self._initial_style = self._fpdf.font_face()
        row = Row(self, style=style, v_align=v_align, min_height=min_height)
        self.rows.append(row)
        for cell in cells:
            if isinstance(cell, dict):
                row.cell(**cell)  # pyright: ignore[reportArgumentType]
            else:
                row.cell(cell)
        return row

    def render(self) -> None:
        "This is an internal method called by `fpdf.FPDF.table()` once the table is finished"
        # Starting with some sanity checks:
        self._cols_count = max(row.cols_count for row in self.rows) if self.rows else 0
        if self._width is None:
            if self._col_widths and isinstance(self._col_widths, NumberClass):
                self._width = self._cols_count * self._col_widths
            else:
                self._width = self._fpdf.epw
        elif self._col_widths and isinstance(self._col_widths, NumberClass):
            if self._cols_count * self._col_widths != self._width:
                raise ValueError(
                    f"Invalid value provided width={self._width} should be a multiple of col_widths={self._col_widths}"
                )
        if self._width > self._fpdf.epw:
            raise ValueError(
                f"Invalid value provided width={self._width}: effective page width is {self._fpdf.epw}"
            )
        if self._table_align == Align.J:
            raise ValueError(
                "JUSTIFY is an invalid value for FPDF.table() 'align' parameter"
            )
        if self._num_heading_rows > 0:
            if not self._headings_style:
                raise ValueError(
                    "headings_style must be provided to FPDF.table() if num_heading_rows>1 or first_row_as_headings=True"
                )
            emphasis = self._headings_style.emphasis
            if emphasis is not None:
                family = self._headings_style.family or self._fpdf.font_family
                font_key = family.lower() + emphasis.style.replace("U", "")
                if font_key not in CORE_FONTS and font_key not in self._fpdf.fonts:
                    # Raising a more explicit error than the one from set_font():
                    raise FPDFException(
                        f"Using font '{family}' with emphasis '{emphasis.style}'"
                        " in table headings require the corresponding font style"
                        " to be added using add_font()"
                    )

        # Defining table global horizontal position:
        prev_x, prev_y, prev_l_margin = self._fpdf.x, self._fpdf.y, self._fpdf.l_margin
        if self._table_align == Align.C:
            self._fpdf.l_margin = (self._fpdf.w - float(self._width)) / 2
            self._fpdf.x = self._fpdf.l_margin
        elif self._table_align == Align.R:
            self._fpdf.l_margin = (
                self._fpdf.w - self._fpdf.r_margin - float(self._width)
            )
            self._fpdf.x = self._fpdf.l_margin
        elif self._fpdf.x != self._fpdf.l_margin:
            self._fpdf.l_margin = self._fpdf.x

        # Pre-Compute the relative x-positions of the individual columns:
        xx = self._fpdf.l_margin + self._outer_border_margin[0]
        cell_x_positions = [xx]
        if self.rows:
            for i in range(self._cols_count):
                xx += self._get_col_width(0, i)
                xx += self._gutter_width
                cell_x_positions.append(xx)

        # Process any rowspans
        rows_info = list(self._compute_rows_info())

        # actually render the cells
        repeat_headings = (
            self._repeat_headings is TableHeadingsDisplay.ON_TOP_OF_EVERY_PAGE
        )
        self._fpdf.y += self._outer_border_margin[1]
        if len(self.rows) > self._num_heading_rows > 0:
            # We avoid having the heading rows alone on a page - issue #1391
            # pylint: disable=protected-access
            self._fpdf._perform_page_break_if_need_be(  # pyright: ignore[reportPrivateUsage]
                sum(
                    rows_info[i].pagebreak_height
                    for i in range(self._num_heading_rows + 1)
                )
            )
        for i in range(len(self.rows)):
            pagebreak_height = rows_info[i].pagebreak_height
            # pylint: disable=protected-access
            page_break = self._fpdf._perform_page_break_if_need_be(  # pyright: ignore[reportPrivateUsage]
                pagebreak_height
            )
            if (
                page_break
                and self._fpdf.y + pagebreak_height > self._fpdf.page_break_trigger
            ):
                # Restoring original position on page:
                self._fpdf.x = prev_x
                self._fpdf.y = prev_y
                self._fpdf.l_margin = prev_l_margin
                raise ValueError(
                    f"The row with index {i} is too high and cannot be rendered on a single page"
                )
            if page_break and repeat_headings and i >= self._num_heading_rows:
                # repeat headings on top:
                self._fpdf.y += self._outer_border_margin[1]
                for row_idx in range(self._num_heading_rows):
                    self._render_table_row(
                        row_idx,
                        rows_info[row_idx],
                        cell_x_positions=cell_x_positions,
                    )
            if i > 0:
                self._fpdf.y += self._gutter_height
            self._render_table_row(i, rows_info[i], cell_x_positions)

        # Restoring altered FPDF settings:
        self._fpdf.l_margin = prev_l_margin
        self._fpdf.x = self._fpdf.l_margin

    def _render_table_row(
        self,
        i: int,
        row_layout_info: "RowLayoutInfo",
        cell_x_positions: Optional[Sequence[float]],
        **kwargs: Any,
    ) -> None:
        row = self.rows[i]
        y = self._fpdf.y  # remember current y position, reset after each cell

        for j, cell in enumerate(row.cells):
            if not isinstance(cell, Cell):
                continue
            self._render_table_cell(
                i,
                j,
                cell,
                row_height=self._line_height,
                cell_height_info=row_layout_info,
                cell_x_positions=cell_x_positions,
                **kwargs,
            )
            self._fpdf.set_y(y)  # restore y position after each cell

        self._fpdf.ln(row_layout_info.height)

    def _render_table_cell(
        self,
        i: int,
        j: int,
        cell: "Cell",
        row_height: float,  # height of a row of text including line spacing
        cell_height_info: Optional[
            "RowLayoutInfo"
        ] = None,  # full height of a cell, including padding, used to render borders and images
        cell_x_positions: Optional[
            Sequence[float]
        ] = None,  # x-positions of the individual columns, pre-calculated for speed. Only relevant when rendering
        **kwargs: Any,
    ) -> Tuple[bool, float, float]:
        # If cell_height_info is provided then we are rendering a cell
        # If cell_height_info is not provided then we are only here to figure out the height of the cell
        #
        # So this function is first called without cell_height_info to figure out the heights of all cells in a row
        # and then called again with cell_height to actually render the cells
        cell_height: Optional[float]
        if cell_height_info is None:
            cell_height = None
            height_query_only = True
        elif cell.rowspan > 1:
            cell_height = cell_height_info.merged_heights[cell.rowspan]
            height_query_only = False
        else:
            cell_height = cell_height_info.height
            height_query_only = False

        page_break_text = False
        page_break_image = False

        # Get style and cell content:

        row = self.rows[i]
        col_width = self._get_col_width(i, j, cell.colspan)
        img_height: float = 0

        text_align = cell.align or self._text_align
        if not isinstance(text_align, (Align, str)):
            text_align = text_align[j]

        style = self._initial_style
        assert style is not None
        cell_mode_fill = self._cell_fill_mode.should_fill_cell(i, j)
        if cell_mode_fill and self._cell_fill_color:
            style = style.replace(fill_color=self._cell_fill_color)
        if i < self._num_heading_rows:
            style = FontFace.combine(style, self._headings_style)
        style = FontFace.combine(style, row.style)
        style = FontFace.combine(style, cell.style)

        padding = Padding.new(cell.padding) if cell.padding else self._padding

        v_align = cell.v_align
        if not v_align:
            v_align = row.v_align
        if not v_align:
            v_align = self._v_align

        # We can not rely on the actual x position of the cell. Notably in case of
        # empty cells or cells with an image only the actual x position is incorrect.
        # Instead, we calculate the x position based on the column widths of the previous columns

        # place cursor (required for images after images)

        # not rendering, cell_x_positions is not relevant (and probably not provided):
        cell_x: float
        if height_query_only:
            cell_x = 0
        else:
            assert cell_x_positions is not None
            cell_x = cell_x_positions[j]
        self._fpdf.set_x(cell_x)

        # render cell border and background
        # if cell_height is defined, that means that we already know the size at which the cell will be rendered
        # so we can draw the borders now
        #
        # If cell_height is None then we're still in the phase of calculating the height of the cell meaning that
        # we do not need to set fonts & draw borders yet.
        if not height_query_only:
            assert cell_height is not None
            x1 = self._fpdf.x
            y1 = self._fpdf.y
            x2 = (
                x1 + col_width
            )  # already includes gutter for cells spanning multiple columns
            y2 = y1 + cell_height

            cell_idx = next(
                i for i, row_cell in enumerate(row.cells) if row_cell is cell
            )
            (
                self._borders_layout.cell_style_getter(
                    row_idx=i,
                    col_idx=sum(1 for cell in row.cells[:cell_idx] if cell is not None),
                    col_pos=j,
                    num_heading_rows=self._num_heading_rows,
                    num_rows=len(self.rows),
                    num_col_idx=sum(1 for cell in row.cells if cell is not None),
                    num_col_pos=row.cols_count,
                )
                .override_cell_border(cell.border)
                .draw_cell_border(
                    self._fpdf,
                    x1,
                    y1,
                    x2,
                    y2,
                    fill_color=style.fill_color if style else None,
                )
            )

            # draw outer box if needed:
            if self._outer_border_width is not None:
                assert self._width is not None
                _remember_linewidth = self._fpdf.line_width
                self._fpdf.set_line_width(float(self._outer_border_width))

                # draw the outer box separated by the gutter dimensions
                # the top and bottom borders are one continuous line
                # whereas the left and right borders are segments because of possible pagebreaks
                x1 = self._fpdf.l_margin
                x2 = x1 + float(self._width)
                y1 = y1 - self._outer_border_margin[1]
                y2 = y2 + self._outer_border_margin[1]

                if j == 0:
                    # lhs border
                    self._fpdf.line(x1, y1, x1, y2)
                if j + cell.colspan == self._cols_count:
                    # rhs border
                    self._fpdf.line(x2, y1, x2, y2)
                    # continuous top line border
                    if i == 0:
                        self._fpdf.line(x1, y1, x2, y1)
                    # continuous bottom line border
                    if i + cell.rowspan == len(self.rows):
                        self._fpdf.line(x1, y2, x2, y2)

                self._fpdf.set_line_width(_remember_linewidth)

        if cell.img:
            x, y = self._fpdf.x, self._fpdf.y

            # if cell_height is None or width is given then call image with h=0
            # calling with h=0 means that the image will be rendered with an auto determined height
            cell_border_line_width = self._fpdf.line_width
            if cell.img_fill_width or cell_height is None:
                img_height = 0
            else:
                img_height = (
                    cell_height - padding.top - padding.bottom - cell_border_line_width
                )

            # apply padding
            self._fpdf.x += padding.left + cell_border_line_width / 2
            self._fpdf.y += padding.top + cell_border_line_width / 2

            image = self._fpdf.image(
                cell.img,
                w=col_width - padding.left - padding.right - cell_border_line_width,
                h=img_height,
                keep_aspect_ratio=True,
                link=cell.link,
            )

            img_height = (
                image.rendered_height
                + padding.top
                + padding.bottom
                + cell_border_line_width
            )

            if img_height + y > self._fpdf.page_break_trigger:
                page_break_image = True

            self._fpdf.set_xy(x, y)

        # render text

        if cell.text:
            dy: float = 0

            if cell_height is not None:
                assert cell_height_info is not None
                actual_text_height = cell_height_info.rendered_heights[j]

                if v_align == VAlign.M:
                    dy = (cell_height - actual_text_height) / 2
                elif v_align == VAlign.B:
                    dy = cell_height - actual_text_height

            self._fpdf.y += dy
            assert style is not None
            with self._fpdf.use_font_face(style):
                page_break_text, cell_height = self._fpdf.multi_cell(  # type: ignore[assignment,misc]
                    w=col_width,
                    h=row_height,
                    text=cell.text,
                    max_line_height=self._line_height,
                    border=0,
                    align=text_align,
                    new_x="RIGHT",
                    new_y="TOP",
                    fill=False,  # fill is already done above
                    markdown=self._markdown,
                    output=MethodReturnValue.PAGE_BREAK | MethodReturnValue.HEIGHT,
                    wrapmode=self._wrapmode,
                    padding=padding,
                    link=cell.link,
                    **kwargs,
                )
                assert isinstance(page_break_text, bool)
                assert isinstance(cell_height, float)

            self._fpdf.y -= dy
        else:
            cell_height = 0

        do_pagebreak: bool = page_break_text or page_break_image
        assert cell_height is not None
        return do_pagebreak, img_height, cell_height

    def _get_col_width(self, i: int, j: int, colspan: int = 1) -> float:
        """Gets width of a column in a table, this excludes the outer gutter (outside the table) but includes the inner gutter
        between columns if the cell spans multiple columns."""

        cols_count = self._cols_count
        assert self._width is not None
        width: float = (
            float(self._width)
            - (cols_count - 1) * self._gutter_width
            - 2 * float(self._outer_border_margin[0])
        )
        gutter_within_cell = max((colspan - 1) * self._gutter_width, 0)

        if not self._col_widths:
            return float(colspan * (width / cols_count) + gutter_within_cell)
        if isinstance(self._col_widths, NumberClass):
            return colspan * self._col_widths + gutter_within_cell
        if j >= len(self._col_widths):
            raise ValueError(
                f"Invalid .col_widths specified: missing width for table() column {j + 1} on row {i + 1}"
            )
        col_width: float = 0
        for k in range(j, j + colspan):
            col_ratio = self._col_widths[k] / sum(self._col_widths)
            col_width += col_ratio * width
            if k != j:
                col_width += self._gutter_width
        return col_width

    def _compute_rows_info(self) -> Iterator["RowLayoutInfo"]:
        # First pass: Regularise the table by processing the rowspan and colspan entries
        active_rowspans: Dict[int, int] = {}
        prev_row_in_col: Dict[int, Optional[Row]] = {}
        for i, row in enumerate(self.rows):
            # Link up rowspans
            active_rowspans, prior_rowspans = row.convert_spans(active_rowspans)
            for col_idx in prior_rowspans:
                # This cell is TableSpan.ROW, so accumulate to the previous row
                prev_row = prev_row_in_col[col_idx]
                if prev_row is not None:
                    # Since Cell objects are frozen, we need to recreate them to update the rowspan
                    cell = prev_row.cells[col_idx]
                    assert isinstance(cell, Cell)
                    prev_row.cells[col_idx] = replace(cell, rowspan=cell.rowspan + 1)
            for j, cell in enumerate(row.cells):
                if isinstance(cell, Cell):
                    # Keep track of the non-span cells
                    prev_row_in_col[j] = row
                    for k in range(j + 1, j + cell.colspan):
                        prev_row_in_col[k] = None
        if len(active_rowspans) != 0:
            raise FPDFException("Rowspan extends beyond end of table")

        # Second pass: Estimate the cell sizes
        rowspan_list: List[RowSpanLayoutInfo] = []
        row_min_heights: List[Number] = []
        row_span_max: List[int] = []
        rendered_heights: List[Dict[int, float]] = []
        # pylint: disable=protected-access
        with self._fpdf._disable_writing():  # pyright: ignore[reportPrivateUsage]
            for i, row in enumerate(self.rows):
                dictated_heights: list[float] = []
                img_heights: list[float] = []
                rendered_heights.append({})

                for j, cell in enumerate(row.cells):
                    if not isinstance(cell, Cell):  # placeholder cell
                        continue

                    # NB: ignore page_break since we might need to assign rowspan padding
                    _, img_height, text_height = self._render_table_cell(
                        i,
                        j,
                        cell,
                        row_height=self._line_height,
                    )
                    if cell.img_fill_width:
                        dictated_height = img_height
                    else:
                        dictated_height = text_height

                    # Store the dictated heights in a dict (not list) because of span elements
                    rendered_heights[i][j] = dictated_height

                    if cell.rowspan > 1:
                        # For spanned rows, use img_height if dictated_height is zero
                        rowspan_list.append(
                            RowSpanLayoutInfo(
                                j, i, cell.rowspan, dictated_height or img_height
                            )
                        )
                        # Often we want rowspans in headings, but issues arise if the span crosses outside the heading
                        is_heading = i < self._num_heading_rows
                        span_outside_heading = i + cell.rowspan > self._num_heading_rows
                        if is_heading and span_outside_heading:
                            raise FPDFException(
                                "Heading includes rowspan beyond the number of heading rows"
                            )
                    else:
                        dictated_heights.append(dictated_height)
                        img_heights.append(img_height)

                # The height of the rows is chosen as follows:
                # The "dictated height" is the space required for text/image, so pick the largest in the row
                # If this is zero, we will fill the space with images, so pick the largest image height
                # If this is still zero (e.g. empty/fully spanned row), use a sensible default
                min_height: float = 0
                if dictated_heights:
                    min_height = max(dictated_heights)
                    if min_height == 0:
                        min_height = max(img_heights)
                if min_height == 0:
                    min_height = self._line_height
                if row.min_height:
                    if min_height < row.min_height:
                        min_height = row.min_height
                elif self._min_row_height:
                    if min_height < self._min_row_height:
                        min_height = float(self._min_row_height)

                row_min_heights.append(min_height)
                row_span_max.append(row.max_rowspan)

        # Sort the spans so we allocate padding to the smallest spans first
        rowspan_list = sorted(rowspan_list, key=lambda span: span.length)

        # Third pass: allocate space required for the rowspans
        row_span_padding: List[float] = [0 for _ in self.rows]
        for span in rowspan_list:
            # accumulate already assigned properties
            max_padding: float = 0
            assigned_height = self._gutter_height * (span.length - 1)
            assigned_padding: float = 0
            for i in span.row_range():
                max_padding = max(max_padding, row_span_padding[i])
                assigned_height += float(row_min_heights[i])
                assigned_padding += row_span_padding[i]

            # does additional padding need to be distributed?
            if assigned_height + assigned_padding < span.contents_height:
                # when there are overlapping rowspans, can we stretch the cells to be evenly padded?
                if span.contents_height > assigned_height + span.length * max_padding:
                    # stretch all cells to have the same padding, for aesthetic reasons
                    padding = (span.contents_height - assigned_height) / span.length
                    for i in span.row_range():
                        row_span_padding[i] = padding
                else:
                    # add proportional padding to the rows
                    extra = span.contents_height - assigned_height - assigned_padding
                    for i in span.row_range():
                        row_span_padding[i] += extra / span.length

        # Fourth pass: compute the final element sizes
        for i, row in enumerate(self.rows):
            row_height: float = float(row_min_heights[i]) + row_span_padding[i]
            # Compute the size of merged cells
            merged_sizes: List[float] = [0, row_height]
            for j in range(i + 1, i + row_span_max[i]):
                merged_sizes.append(
                    merged_sizes[-1]
                    + self._gutter_height
                    + float(row_min_heights[j])
                    + row_span_padding[j]
                )
            # Pagebreak should not occur within ANY rowspan, so validate ACCUMULATED rowspans
            # This gets complicated because of overlapping rowspans (see `test_table_with_rowspan_and_pgbreak()`)
            # Eventually, this should be refactored to rearrange cells to permit breaks within spans
            pagebreak_height = row_height
            pagebreak_row = i + row_span_max[i]
            j = i + 1
            while j < pagebreak_row:
                # NB: this can't be a for loop because the upper limit might keep changing
                pagebreak_row = max(pagebreak_row, j + row_span_max[j])
                pagebreak_height += (
                    self._gutter_height
                    + float(row_min_heights[j])
                    + row_span_padding[j]
                )
                j += 1

            yield RowLayoutInfo(
                merged_sizes[1], pagebreak_height, rendered_heights[i], merged_sizes
            )


class Row:
    "Object that `Table.row()` yields, used to build a row in a table"

    def __init__(
        self,
        table: Table,
        style: Optional[FontFace] = None,
        v_align: Optional[str | VAlign] = None,
        min_height: Optional[float] = None,
    ) -> None:
        self._table = table
        self.cells: List[Optional[TableSpan | Cell]] = []
        self.style: Optional[FontFace] = style
        self.v_align = VAlign.coerce(v_align) if v_align else v_align
        self.min_height = min_height

    @property
    def cols_count(self) -> int:
        return sum(getattr(cell, "colspan", cell is not None) for cell in self.cells)

    @property
    def max_rowspan(self) -> int:
        spans = {cell.rowspan for cell in self.cells if isinstance(cell, Cell)}
        return max(spans) if len(spans) else 1

    def convert_spans(
        self, active_rowspans: Dict[int, int]
    ) -> Tuple[Dict[int, int], Sequence[int]]:
        # convert colspans
        prev_col = 0
        cells: List[Optional[Cell | TableSpan]] = []
        for i, cell in enumerate(self.cells):
            if cell is None:
                continue
            if cell == TableSpan.COL:
                prev_cell = cells[prev_col]
                if not isinstance(prev_cell, Cell):
                    raise FPDFException(
                        "Invalid location for TableSpan.COL placeholder entry"
                    )
                cells[prev_col] = replace(prev_cell, colspan=prev_cell.colspan + 1)
                cells.append(None)  # processed
            else:
                cells.append(cell)
                prev_col = i
                if isinstance(cell, Cell) and cell.colspan > 1:  # expand any colspans
                    cells.extend([None] * (cell.colspan - 1))
        # now we can correctly interpret active_rowspans
        remaining_rowspans: dict[int, int] = {}
        for k, v in active_rowspans.items():
            cells.insert(k, None)
            if v > 1:
                remaining_rowspans[k] = v - 1
        # accumulate any rowspans
        reverse_rowspans: list[int] = []
        for i, cell in enumerate(cells):
            if isinstance(cell, Cell) and cell.rowspan > 1:
                for k in range(i, i + cell.colspan):
                    remaining_rowspans[k] = cell.rowspan - 1
            elif cell == TableSpan.ROW:
                reverse_rowspans.append(i)
                cells[i] = None  # processed
        self.cells = cells
        return remaining_rowspans, reverse_rowspans

    def cell(
        self,
        text: str = "",
        align: Optional[str | Align] = None,
        v_align: Optional[str | VAlign] = None,
        style: Optional[FontFace] = None,
        img: Optional[ImageType] = None,
        img_fill_width: bool = False,
        colspan: int = 1,
        rowspan: int = 1,
        padding: Optional[Tuple[float, ...]] = None,
        link: Optional[str | int] = None,
        border: CellBordersLayout = CellBordersLayout.INHERIT,
    ) -> Union["Cell", TableSpan]:
        """
        Adds a cell to the row.

        Args:
            text (str): string content, can contain several lines.
                In that case, the row height will grow proportionally.
            align (str, fpdf.enums.Align): optional text alignment.
            v_align (str, fpdf.enums.VAlign): optional vertical text alignment.
            style (fpdf.fonts.FontFace): optional text style.
            img: optional. Either a string representing a file path to an image,
                an URL to an image, an io.BytesIO, or a instance of `PIL.Image.Image`.
            img_fill_width (bool): optional, defaults to False. Indicates to render the image
                using the full width of the current table column.
            colspan (int): optional number of columns this cell should span.
            rowspan (int): optional number of rows this cell should span.
            padding (tuple): optional padding (left, top, right, bottom) for the cell.
            link (str, int): optional link, either an URL or an integer returned by `FPDF.add_link`, defining an internal link to a page
            border (fpdf.enums.CellBordersLayout): optional cell borders, defaults to `CellBordersLayout.INHERIT`

        """
        if text and img:
            raise NotImplementedError(
                "fpdf2 currently does not support inserting text with an image in the same table cell."
                " Pull Requests are welcome to implement this 😊"
            )

        if isinstance(text, TableSpan):
            # Special placeholder object, converted to colspan/rowspan during processing
            self.cells.append(text)
            return text

        if not style:
            # pylint: disable=protected-access
            # We capture the current font settings:
            font_face = (
                self._table._fpdf.font_face()  # pyright: ignore[reportPrivateUsage]
            )
            if font_face not in (
                self.style,
                self._table._initial_style,  # pyright: ignore[reportPrivateUsage]
            ):
                style = font_face

        cell = Cell(
            text,
            align,
            VAlign.coerce(v_align) if v_align else v_align,
            style,
            img,
            img_fill_width,
            colspan,
            rowspan,
            padding,
            link,
            CellBordersLayout.coerce(border),
        )
        self.cells.append(cell)
        return cell


@dataclass(frozen=True)
class Cell:
    "Internal representation of a table cell"

    __slots__ = (  # RAM usage optimization
        "text",
        "align",
        "v_align",
        "style",
        "img",
        "img_fill_width",
        "colspan",
        "rowspan",
        "padding",
        "link",
        "border",
    )
    text: str
    align: Optional[Union[str, Align]]
    v_align: Optional[Union[str, VAlign]]
    style: Optional[FontFace]
    img: Optional[ImageType]
    img_fill_width: bool
    colspan: int
    rowspan: int
    padding: Optional[float | Sequence[float] | Padding]
    link: Optional[Union[str, int]]
    border: CellBordersLayout

    def write(self, text: str, align: Optional[Align | str] = None) -> None:
        raise NotImplementedError("Not implemented yet")


@dataclass(frozen=True)
class RowLayoutInfo:
    height: float
    # accumulated rowspans to take in account when considering page breaks:
    pagebreak_height: float
    # heights of every cell in the row:
    rendered_heights: Dict[int, float]
    merged_heights: List[float]


@dataclass(frozen=True)
class RowSpanLayoutInfo:
    column: int
    start: int
    length: int
    contents_height: float

    def row_range(self) -> range:
        return range(self.start, self.start + self.length)


def draw_box_borders(
    pdf: "FPDF",
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    border: str | Literal[0, 1],
    fill_color: Optional[DeviceCMYK | DeviceGray | DeviceRGB] = None,
) -> None:
    """Draws a box using the provided style - private helper used by table for drawing the cell and table borders.
    Difference between this and rect() is that border can be defined as "L,R,T,B" to draw only some of the four borders;
    compatible with get_border(i,k)

    See Also: rect()"""
    if fill_color:
        prev_fill_color = pdf.fill_color
        pdf.set_fill_color(fill_color)

    sl: list[str] = []

    k = pdf.k

    # y top to bottom instead of bottom to top
    y1 = pdf.h - y1
    y2 = pdf.h - y2

    # scale
    x1 *= k
    x2 *= k
    y2 *= k
    y1 *= k

    if isinstance(border, str) and set(border).issuperset("LTRB"):
        border = 1

    if fill_color:
        op = "B" if border == 1 else "f"
        sl.append(f"{x1:.2f} {y2:.2f} " f"{x2 - x1:.2f} {y1 - y2:.2f} re {op}")
    elif border == 1:
        sl.append(f"{x1:.2f} {y2:.2f} " f"{x2 - x1:.2f} {y1 - y2:.2f} re S")

    if isinstance(border, str):
        if "L" in border:
            sl.append(f"{x1:.2f} {y2:.2f} m " f"{x1:.2f} {y1:.2f} l S")
        if "B" in border:
            sl.append(f"{x1:.2f} {y2:.2f} m " f"{x2:.2f} {y2:.2f} l S")
        if "R" in border:
            sl.append(f"{x2:.2f} {y2:.2f} m " f"{x2:.2f} {y1:.2f} l S")
        if "T" in border:
            sl.append(f"{x1:.2f} {y1:.2f} m " f"{x2:.2f} {y1:.2f} l S")

    s = " ".join(sl)
    pdf._out(  # pyright: ignore[reportPrivateUsage] # pylint: disable=protected-access
        s
    )

    if fill_color:
        pdf.set_fill_color(prev_fill_color)  # type: ignore[arg-type]



from pathlib import Path

import fpdf
from fpdf.enums import StrokeJoinStyle
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_regular_polygon(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    draw_regular_polygon(pdf)
    assert_pdf_equal(pdf, HERE / "regular_polygon.pdf", tmp_path)


def test_regular_polygon_with_round_joins(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    with pdf.local_context(stroke_join_style=StrokeJoinStyle.ROUND):
        draw_regular_polygon(pdf)
    assert_pdf_equal(pdf, HERE / "regular_polygon_with_round_joins.pdf", tmp_path)


def test_regular_polygon_with_bevel_joins(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    with pdf.local_context(stroke_join_style=StrokeJoinStyle.BEVEL):
        draw_regular_polygon(pdf)
    assert_pdf_equal(pdf, HERE / "regular_polygon_with_bevel_joins.pdf", tmp_path)


def draw_regular_polygon(pdf):
    # no fill with line
    pdf.regular_polygon(10, 35, 3, 25)
    pdf.regular_polygon(40, 35, 4, 25)
    pdf.regular_polygon(70, 35, 5, 25)
    pdf.regular_polygon(100, 35, 6, 25)
    pdf.regular_polygon(130, 35, 7, 25)
    pdf.regular_polygon(160, 35, 8, 25)

    # fill and color test
    pdf.set_fill_color(134, 200, 15)
    pdf.regular_polygon(10, 75, 3, 25, style="df")
    pdf.regular_polygon(40, 75, 4, 25, style="df")
    pdf.regular_polygon(70, 75, 5, 25, style="df")
    pdf.regular_polygon(100, 75, 6, 25, style="df")
    pdf.regular_polygon(130, 75, 7, 25, style="df")
    pdf.regular_polygon(160, 75, 8, 25, style="df")

    # draw color test
    pdf.set_fill_color(0, 0, 0)
    pdf.set_draw_color(204, 0, 204)
    pdf.regular_polygon(10, 115, 3, 25)
    pdf.regular_polygon(40, 115, 4, 25)
    pdf.regular_polygon(70, 115, 5, 25)
    pdf.regular_polygon(100, 115, 6, 25)
    pdf.regular_polygon(130, 115, 7, 25)
    pdf.regular_polygon(160, 115, 8, 25)

    # line width test
    pdf.set_line_width(1)
    pdf.regular_polygon(10, 155, 3, 25)
    pdf.regular_polygon(40, 155, 4, 25)
    pdf.regular_polygon(70, 155, 5, 25)
    pdf.regular_polygon(100, 155, 6, 25)
    pdf.regular_polygon(130, 155, 7, 25)
    pdf.regular_polygon(160, 155, 8, 25)

    # line color and fill color
    pdf.set_fill_color(3, 190, 252)
    pdf.regular_polygon(10, 195, 3, 25, style="df")
    pdf.regular_polygon(40, 195, 4, 25, style="df")
    pdf.regular_polygon(70, 195, 5, 25, style="df")
    pdf.regular_polygon(100, 195, 6, 25, style="df")
    pdf.regular_polygon(130, 195, 7, 25, style="df")
    pdf.regular_polygon(160, 195, 8, 25, style="df")

    # rotation test
    pdf.set_draw_color(0, 0, 255)
    pdf.regular_polygon(10, 235, 3, 25, 30)
    pdf.regular_polygon(40, 235, 4, 25, 45)
    pdf.regular_polygon(70, 235, 5, 25, 200)
    pdf.regular_polygon(100, 235, 6, 25, 0)
    pdf.regular_polygon(130, 235, 7, 25, 13)
    pdf.regular_polygon(160, 235, 8, 25, 22.5)

    # fill only
    pdf.regular_polygon(10, 275, 3, 25, style="f")
    pdf.regular_polygon(40, 275, 4, 25, style="f")
    pdf.regular_polygon(70, 275, 5, 25, style="f")
    pdf.regular_polygon(100, 275, 6, 25, style="f")
    pdf.regular_polygon(130, 275, 7, 25, style="f")
    pdf.regular_polygon(160, 275, 8, 25, style="f")



import math
from pathlib import Path

from fpdf import FPDF
from fpdf.enums import Angle
from test.conftest import assert_pdf_equal, LOREM_IPSUM

HERE = Path(__file__).resolve().parent


def draw_mirror_line(pdf, origin, angle):
    """
    A helper method which converts a given angle and origin to two co-ordinates to
    then draw a line.
    Used to help visualize & test mirror transformations.

    Args:
        pdf (fpdf.FPDF): pdf to modify
        origin (float, Sequence[float, float]): a point on the mirror line
        angle: (fpdf.enums.Angle): the direction of the mirror line
    """
    x, y = origin
    try:
        theta = Angle.coerce(angle).value
    except ValueError:
        theta = angle

    cos_theta, sin_theta = (
        math.cos(math.radians(theta)),
        math.sin(math.radians(theta)) * -1,
    )

    x1 = x - (150 * cos_theta)
    y1 = y - (150 * sin_theta)
    x2 = x + (150 * cos_theta)
    y2 = y + (150 * sin_theta)
    pdf.line(x1=x1, y1=y1, x2=x2, y2=y2)


def test_mirror(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_line_width(0.5)
    pdf.set_draw_color(r=255, g=128, b=0)

    img_filepath = HERE / "image/png_images/66ac49ef3f48ac9482049e1ab57a53e9.png"

    pdf.image(img_filepath, x=100, y=100)

    pdf.image(img_filepath, x=100, y=100)
    with pdf.mirror((pdf.epw / 2, pdf.eph / 2.5), "WEST"):
        draw_mirror_line(pdf, (pdf.epw / 2, pdf.eph / 2.5), "WEST")
        pdf.image(img_filepath, x=100, y=100)

    with pdf.mirror((pdf.epw / 2.5, pdf.eph / 2), "SOUTH"):
        pdf.set_draw_color(r=128, g=0, b=0)
        draw_mirror_line(pdf, (pdf.epw / 2.5, pdf.eph / 2), "SOUTH")
        pdf.image(img_filepath, x=100, y=100)

    with pdf.mirror((pdf.epw / 1.5, pdf.eph / 1.5), "SOUTHWEST"):
        pdf.set_draw_color(r=0, g=0, b=128)
        draw_mirror_line(pdf, (pdf.epw / 1.5, pdf.eph / 1.5), "SOUTHWEST")
        pdf.image(img_filepath, x=100, y=100)

    with pdf.mirror((pdf.epw / 3, pdf.eph / 2.5), "SOUTHEAST"):
        pdf.set_draw_color(r=0, g=128, b=0)
        draw_mirror_line(pdf, (pdf.epw / 3, pdf.eph / 2.5), "SOUTHEAST")
        pdf.image(img_filepath, x=100, y=100)

    assert_pdf_equal(pdf, HERE / "mirror.pdf", tmp_path)


def test_mirror_with_angle_as_number(tmp_path):
    pdf = FPDF()
    pdf.set_font("helvetica", size=50)
    pdf.add_page()
    x, y = 50, 50
    pdf.text(x, y, text="mirror this text")
    with pdf.mirror((x, y), 180):
        pdf.set_text_color(r=255, g=128, b=0)
        pdf.text(x, y, text="mirror this text")
    assert_pdf_equal(pdf, HERE / "mirror_with_angle_as_number.pdf", tmp_path)


def test_mirror_text(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=12)

    pdf.text(pdf.epw / 2, pdf.epw / 2, text="mirror this text")

    with pdf.mirror((pdf.epw / 2, pdf.eph / 2.5), "EAST"):
        pdf.text(pdf.epw / 2, pdf.eph / 2, text="mirrored text E/W")

    with pdf.mirror((pdf.epw / 2.5, pdf.eph / 2), "NORTH"):
        pdf.text(pdf.epw / 2, pdf.eph / 2, text="mirrored text N/S")

    with pdf.mirror((pdf.epw / 1.5, pdf.eph / 1.5), "NORTHWEST"):
        pdf.text(pdf.epw / 2, pdf.eph / 2, text="mirrored text NW/SE")

    with pdf.mirror((pdf.epw / 2.5, pdf.eph / 2.5), "NORTHEAST"):
        pdf.text(pdf.epw / 2, pdf.eph / 2, text="mirrored text NE/SW")

    assert_pdf_equal(pdf, HERE / "mirror_text.pdf", tmp_path)


def test_mirror_cell(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    pdf.set_fill_color(255, 255, 0)

    pdf.cell(text="cell to be mirrored repeatedly")
    pdf.x = pdf.l_margin
    with pdf.mirror((pdf.epw / 2, 0), "NORTH"):
        draw_mirror_line(pdf, (pdf.epw / 2, 0), "NORTH")
        pdf.cell(text="cell mirrored", fill=True)
        pdf.cell(text="cell mirrored", fill=True)
        pdf.cell(text="cell mirrored", fill=True)
        pdf.ln(40)

    pdf.cell(text="cell text 1")
    pdf.x = pdf.l_margin
    with pdf.mirror((pdf.epw / 2, pdf.eph / 4), "EAST"):
        draw_mirror_line(pdf, (pdf.epw / 2, pdf.eph / 4), "EAST")
        pdf.cell(text="cell text 1", fill=True)
        pdf.ln(40)

    pdf.cell(text="cell text 2")
    pdf.x = pdf.l_margin
    with pdf.mirror((pdf.epw / 2, 0), "SOUTHWEST"):
        draw_mirror_line(pdf, (pdf.epw / 2, 0), "SOUTHWEST")
        pdf.cell(text="cell text 2", fill=True)
        pdf.ln(40)

    pdf.cell(text="cell text 3")
    pdf.x = pdf.l_margin
    with pdf.mirror((pdf.epw / 2, pdf.eph / 4), "NORTHEAST"):
        draw_mirror_line(pdf, (pdf.epw / 2, pdf.eph / 4), "NORTHEAST")
        pdf.cell(text="cell text 3", fill=True, border=1)
        pdf.ln(40)

    assert_pdf_equal(pdf, HERE / "mirror_cell.pdf", tmp_path)


def test_mirror_multi_cell(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    pdf.set_fill_color(255, 255, 0)

    pdf.multi_cell(w=50, text=LOREM_IPSUM[:200])

    pdf.x = pdf.l_margin
    pdf.y = pdf.t_margin
    with pdf.mirror((pdf.epw / 2, pdf.eph / 4), "NORTHEAST"):
        draw_mirror_line(pdf, (pdf.epw / 2, pdf.eph / 4), "NORTHEAST")
        pdf.multi_cell(w=50, text=LOREM_IPSUM[:200], fill=True)
        pdf.ln(20)

    prev_y = pdf.y
    pdf.multi_cell(w=100, text=LOREM_IPSUM[:200])
    pdf.x = pdf.l_margin
    pdf.y = prev_y

    with pdf.mirror((0, pdf.eph / 2), "EAST"):
        draw_mirror_line(pdf, (0, pdf.eph / 2), "EAST")
        pdf.multi_cell(w=100, text=LOREM_IPSUM[:200], fill=True)
        pdf.ln(150)

    prev_y = pdf.y
    pdf.multi_cell(w=120, text=LOREM_IPSUM[:120])
    pdf.x = pdf.l_margin
    pdf.y = prev_y

    with pdf.mirror((pdf.epw / 2, pdf.eph), "SOUTH"):
        draw_mirror_line(pdf, (pdf.epw / 2, pdf.eph), "SOUTH")
        pdf.multi_cell(w=120, text=LOREM_IPSUM[:120], fill=True, border=1)

    assert_pdf_equal(pdf, HERE / "mirror_multi_cell.pdf", tmp_path)



from pathlib import Path

from fpdf import FPDF
from fpdf.enums import XPos, YPos
from test.conftest import assert_pdf_equal

import pytest

HERE = Path(__file__).resolve().parent


def test_named_destinations(tmp_path):
    """Test PDF generation with named destinations and links.

    This test demonstrates how to create a hierarchical structure of named destinations
    by using named destinations directly and binding them to pages later.
    """
    # Create PDF
    pdf = FPDF()

    # Create the first page first
    pdf.add_page()

    # Create the main TOC destination directly
    pdf.set_link(name="dests/main")

    # For backward compatibility, you can still use add_link with named destinations
    # We only need to create a link for subsection1 this way for testing
    sub1_link = pdf.add_link(name="dests/subsection1", page=2)

    # Subsections 2 and 3 will be referenced directly by name
    # without pre-creating them

    # First page is already created
    pdf.set_font("Helvetica", size=16)
    # Main TOC destination is already bound to this location
    pdf.cell(0, 10, "Table of Contents", new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.ln(10)
    pdf.set_font("Helvetica", size=12)

    # Add links to subsections in the TOC
    pdf.cell(
        0,
        10,
        "Subsection 1: Concepts",
        link=sub1_link,
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )

    # You can directly reference a destination by name
    # Even if it hasn't been set yet
    pdf.cell(
        0,
        10,
        "Subsection 2: Implementation",
        link="#dests/subsection2",  # Using the '#' prefix to reference the named destination
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )

    # Use direct named reference for subsection 3 too
    pdf.cell(
        0,
        10,
        "Subsection 3: Advanced Usage",
        link="#dests/subsection3",  # Named destinations automatically created when referenced
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )

    # Subsection 1 page
    pdf.add_page()
    pdf.set_font("Helvetica", size=16)
    # Bind subsection1 destination to this location
    pdf.set_link(sub1_link)
    pdf.cell(0, 10, "Subsection 1: Concepts", new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.set_font("Helvetica", size=12)
    pdf.multi_cell(
        0,
        10,
        (
            "This page demonstrates binding a previously created named destination. "
            "The destination 'dests/subsection1' was created before this page existed "
            "and is now bound to this location.\n\n"
            "Click on the table of contents link to return."
        ),
    )
    # Add a link back to TOC
    pdf.cell(
        0,
        10,
        "Return to Table of Contents",
        link="#dests/main",
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )

    # Subsection 2 page
    pdf.add_page()
    pdf.set_font("Helvetica", size=16)
    # Bind subsection2 destination to this location using name directly
    pdf.set_link(name="dests/subsection2")
    pdf.cell(0, 10, "Subsection 2: Implementation", new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.set_font("Helvetica", size=12)
    pdf.multi_cell(
        0,
        10,
        (
            "This page shows how named destinations maintain stability when page numbers "
            "change. Even if pages are added or removed, the named destination 'dests/subsection2' "
            "will always point to this content.\n\n"
            "Click on the table of contents link to return."
        ),
    )
    # Add a link back to TOC
    pdf.cell(
        0,
        10,
        "Return to Table of Contents",
        link="#dests/main",
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )

    # Subsection 3 page
    pdf.add_page()
    pdf.set_font("Helvetica", size=16)
    # Bind subsection3 destination to this location using name directly
    pdf.set_link(name="dests/subsection3")
    pdf.cell(0, 10, "Subsection 3: Advanced Usage", new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.set_font("Helvetica", size=12)
    pdf.multi_cell(
        0,
        10,
        (
            "Named destinations can be referenced anywhere in the document, including in "
            "bookmarks, outlines, and from external documents. The hierarchical naming scheme "
            "'dests/subsection3' allows for organized structure.\n\n"
            "Click on the table of contents link to return."
        ),
    )
    # Add a link back to TOC using direct named reference
    pdf.cell(
        0,
        10,
        "Return to Table of Contents",
        link="#dests/main",
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )

    # Compare generated PDF to reference using assert_pdf_equal
    assert_pdf_equal(pdf, HERE / "named_destinations.pdf", tmp_path)


def test_invalid_destination():
    """Test that using non-existent index still raises a KeyError."""
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(KeyError):
        pdf.set_link(99999)


def test_duplicate_destinations():
    """Test that creating multiple destinations with the same name works."""
    pdf = FPDF()
    pdf.add_page()
    link1 = pdf.add_link(name="test")
    link2 = pdf.add_link(name="test")
    # Should not raise any error when setting either link
    pdf.set_link(link1)
    pdf.set_link(link2)


def test_set_link_with_name():
    """Test setting a named destination using set_link with name parameter."""
    pdf = FPDF()
    pdf.add_page()

    # Create a named destination with set_link
    name = pdf.set_link(name="direct_destination")

    # Verify the name was returned
    assert name == "direct_destination"

    # Verify the destination exists
    assert "direct_destination" in pdf.named_destinations

    # Create another page and reference the destination
    pdf.add_page()
    pdf.set_font("Helvetica", size=12)
    pdf.cell(0, 10, "Link to first page", link="#direct_destination")

    # This should not raise an error
    dest_link = pdf.get_named_destination("direct_destination")
    assert dest_link is not None


def test_unset_destinations_error():
    """Test that referencing but not setting a destination raises an error."""
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=12)

    # Reference a destination but don't set it
    pdf.cell(0, 10, "Link to nowhere", link="#missing-destination")

    # Trying to output should raise an exception
    with pytest.raises(Exception) as excinfo:
        pdf.output(HERE / "should_not_be_created.pdf")

    # Verify error message contains the missing destination name
    assert "missing-destination" in str(excinfo.value)



# pylint: disable=no-self-use, protected-access, redefined-outer-name, unused-argument

import copy
import math
from pathlib import Path
import re

from fpdf.enums import (
    BlendMode,
    IntersectionRule,
    PathPaintRule,
    StrokeCapStyle,
    StrokeJoinStyle,
)
from fpdf.errors import FPDFException
from fpdf.fpdf import FPDF
from fpdf.output import ResourceCatalog
from fpdf.syntax import Name
from test.conftest import assert_pdf_equal
from fpdf.drawing import (
    ClippingPath,
    DrawingContext,
    GraphicsContext,
    GraphicsStyle,
    ImplicitClose,
    Line,
    Move,
    PaintedPath,
    RelativeLine,
    render_pdf_primitive,
)
from fpdf.drawing_primitives import (
    DeviceCMYK,
    DeviceGray,
    DeviceRGB,
    Point,
    Transform,
    check_range,
    cmyk8,
    color_from_hex_string,
    color_from_rgb_string,
    gray8,
    number_to_str,
    rgb8,
)

import pytest

from . import parameters

HERE = Path(__file__).resolve().parent
bad_path_chars = re.compile(r"[\[\]=#, ]")


@pytest.fixture(scope="function")
def auto_pdf(request):
    pdf = FPDF(unit="mm", format=(10, 10))
    pdf.add_page()

    yield pdf


@pytest.fixture(scope="function")
def auto_pdf_cmp(request, tmp_path, auto_pdf):
    yield auto_pdf

    assert_pdf_equal(
        auto_pdf,
        HERE / "generated_pdf" / f"{bad_path_chars.sub('_', request.node.name)}.pdf",
        tmp_path,
    )


@pytest.fixture
def open_path_drawing():
    path = PaintedPath()
    path.move_to(2, 2)
    path.curve_to(4, 3, 6, 3, 8, 2)
    path.line_to(8, 8)
    path.curve_to(6, 7, 4, 7, 2, 8)

    path.style.fill_color = "#CCCF"
    path.style.stroke_color = "#333F"
    path.style.stroke_width = 0.5
    path.style.stroke_cap_style = "round"
    path.style.stroke_join_style = "round"
    path.style.auto_close = False

    return path


@pytest.fixture
def prepared_blend_path(open_path_drawing):
    open_path_drawing.style.fill_color = rgb8(100, 120, 200)
    open_path_drawing.style.stroke_color = rgb8(200, 120, 100)
    open_path_drawing.style.fill_opacity = 0.8
    open_path_drawing.style.stroke_opacity = 0.8

    return open_path_drawing


class TestUtilities:
    def test_range_check(self):
        with pytest.raises(ValueError):
            check_range(-1, minimum=0.0, maximum=0.0)

        check_range(0, minimum=0.0, maximum=1.0)
        check_range(1, minimum=0.0, maximum=1.0)

    @pytest.mark.parametrize("number, converted", parameters.numbers)
    def test_number_to_str(self, number, converted):
        assert number_to_str(number) == converted

    @pytest.mark.parametrize("primitive, result", parameters.pdf_primitives)
    def test_render_primitive(self, primitive, result):
        assert render_pdf_primitive(primitive) == result

    # Add check for bad primitives: class without serialize, dict with non-Name
    # keys. Check for proper escaping of Name and string edge cases
    @pytest.mark.parametrize("primitive, error_type", parameters.pdf_bad_primitives)
    def test_error_on_bad_primitive(self, primitive, error_type):
        with pytest.raises(error_type):
            render_pdf_primitive(primitive)


class TestGraphicsStateDictRegistry:
    @pytest.fixture
    def new_style_registry(self):
        return ResourceCatalog()

    def test_empty_style(self, new_style_registry):
        style = GraphicsStyle()

        result = new_style_registry.register_graphics_style(style)

        assert result is None

    def test_adding_styles(self, new_style_registry):
        first = GraphicsStyle()
        second = GraphicsStyle()

        first.stroke_width = 1
        second.stroke_width = 2

        first_name = new_style_registry.register_graphics_style(first)
        second_name = new_style_registry.register_graphics_style(second)

        assert isinstance(first_name, Name)
        assert isinstance(second_name, Name)
        assert first_name != second_name
        # I don't think it's necessary to check that first_name is actually e.g. "GS1", since
        # the naming pattern is an implementation detail that can change without
        # affecting anything.

    def test_style_deduplication(self, new_style_registry):
        first = GraphicsStyle()
        second = GraphicsStyle()

        first.stroke_width = 1
        second.stroke_width = 1

        first_name = new_style_registry.register_graphics_style(first)
        second_name = new_style_registry.register_graphics_style(second)

        assert isinstance(first_name, Name)
        assert isinstance(second_name, Name)
        assert first_name == second_name


class TestColors:
    def test_device_rgb(self):
        rgb = DeviceRGB(r=1, g=0.5, b=0.25)
        rgba = DeviceRGB(r=1, g=0.5, b=0.25, a=0.75)

        assert rgb.colors == (1, 0.5, 0.25)
        assert rgba.colors == (1, 0.5, 0.25)

        with pytest.raises(ValueError):
            DeviceRGB(r=2, g=1, b=0)

    def test_device_gray(self):
        gray = DeviceGray(g=0.5)
        gray_a = DeviceGray(g=0.5, a=0.75)

        assert gray.colors == (0.5, 0.5, 0.5)
        assert gray_a.colors == (0.5, 0.5, 0.5)

        with pytest.raises(ValueError):
            DeviceGray(g=2)

    def test_device_cmyk(self):
        cmyk = DeviceCMYK(c=1, m=1, y=1, k=0)
        cmyk_a = DeviceCMYK(c=1, m=1, y=1, k=0, a=0.5)

        assert cmyk.colors == (1, 1, 1, 0)
        assert cmyk_a.colors == (1, 1, 1, 0)

        assert cmyk.serialize() == "1 1 1 0 k"
        assert cmyk_a.serialize() == "1 1 1 0 k"

        with pytest.raises(ValueError):
            DeviceCMYK(c=2, m=1, y=1, k=0)

    def test_rgb8(self):
        rgb = rgb8(255, 254, 253)
        rgba = rgb8(r=255, g=254, b=253, a=252)

        assert rgb == DeviceRGB(r=1.0, g=254 / 255, b=253 / 255)
        assert rgba == DeviceRGB(r=1.0, g=254 / 255, b=253 / 255, a=252 / 255)

    def test_gray8(self):
        gray = gray8(255)
        gray_a = gray8(g=255, a=254)

        assert gray == DeviceGray(g=1.0)
        assert gray_a == DeviceGray(g=1.0, a=254 / 255)

    def test_cmyk8(self):
        cmyk = cmyk8(255, 254, 253, 252)
        cmyk_a = cmyk8(c=255, m=254, y=253, k=252, a=251)

        assert cmyk == DeviceCMYK(c=1.0, m=254 / 255, y=253 / 255, k=252 / 255)
        assert cmyk_a == DeviceCMYK(
            c=1.0, m=254 / 255, y=253 / 255, k=252 / 255, a=251 / 255
        )

    @pytest.mark.parametrize("hex_string, result", parameters.hex_colors)
    def test_hex_string_parser(self, hex_string, result):
        if isinstance(result, type) and issubclass(result, Exception):
            with pytest.raises(result):
                color_from_hex_string(hex_string)
        else:
            assert color_from_hex_string(hex_string) == result

    @pytest.mark.parametrize("rgb_string, result", parameters.rgb_colors)
    def test_rgb_string_parser(self, rgb_string, result):
        if isinstance(result, type) and issubclass(result, Exception):
            with pytest.raises(result):
                color_from_rgb_string(rgb_string)
        else:
            assert color_from_rgb_string(rgb_string) == result


class TestPoint:
    def test_render(self):
        pt = Point(1, 1)

        assert pt.render() == "1 1"

    def test_dot(self):
        pt1 = Point(1, 2)
        pt2 = Point(3, 4)

        assert pt1.dot(pt2) == 11
        assert pt2.dot(pt1) == 11

        with pytest.raises(TypeError):
            pt1.dot(1)

    def test_angle(self):
        pt1 = Point(1, 0)
        pt2 = Point(0, 1)

        assert pt1.angle(pt2) == pytest.approx(math.pi / 2)
        assert pt2.angle(pt1) == pytest.approx(-(math.pi / 2))

        with pytest.raises(TypeError):
            pt1.angle(1)

    def test_magnitude(self):
        pt1 = Point(2, 0)

        assert pt1.mag() == 2.0

    def test_add(self):
        pt1 = Point(1, 0)
        pt2 = Point(0, 1)

        assert pt1 + pt2 == Point(1, 1)
        assert pt2 + pt1 == Point(1, 1)

        with pytest.raises(TypeError):
            _ = pt1 + 2

    def test_sub(self):
        pt1 = Point(1, 0)
        pt2 = Point(0, 1)

        assert pt1 - pt2 == Point(1, -1)
        assert pt2 - pt1 == Point(-1, 1)

        with pytest.raises(TypeError):
            _ = pt1 - 2

    def test_neg(self):
        pt = Point(1, 2)

        assert -pt == Point(-1, -2)

    def test_mul(self):
        pt = Point(1, 2)

        assert pt * 2 == Point(2, 4)
        assert 2 * pt == Point(2, 4)

        with pytest.raises(TypeError):
            _ = pt * pt

    def test_div(self):
        pt = Point(1, 2)

        assert (pt / 2) == Point(0.5, 1.0)
        assert (pt // 2) == Point(0, 1)

        with pytest.raises(TypeError):
            _ = 2 / pt

        with pytest.raises(TypeError):
            _ = 2 // pt

        with pytest.raises(TypeError):
            _ = pt / pt

        with pytest.raises(TypeError):
            _ = pt // pt

    def test_matmul(self):
        pt = Point(2, 4)
        tf = Transform.translation(2, 2)

        assert pt @ tf == Point(4, 6)

        with pytest.raises(TypeError):
            _ = tf @ pt

        with pytest.raises(TypeError):
            _ = pt @ pt

    def test_str(self):
        pt = Point(2, 4)

        assert str(pt) == "(x=2, y=4)"


class TestTransform:
    @pytest.mark.parametrize("tf, pt, result", parameters.transforms)
    def test_constructors(self, tf, pt, result):
        # use of pytest.approx here works because Point is a NamedTuple
        # (approx supports tuples) and wouldn't work if it were a normal object
        assert pt @ tf == pytest.approx(result)

    def test_chaining(self):
        tf = Transform.identity().translate(1, 1).scale(2, 1).rotate(0.3).shear(0, 0.5)
        assert tf == pytest.approx(
            Transform(
                a=1.910672978251212,
                b=1.546376902448285,
                c=-0.29552020666133955,
                d=0.8075763857949362,
                e=1.6151527715898724,
                f=2.353953288243221,
            )
        )

    def test_about(self):
        tf = Transform.scaling(2).about(10, 10)

        assert tf == Transform(a=2, b=0, c=0, d=2, e=-10, f=-10)
        assert tf == Transform.translation(-10, -10).scale(2).translate(10, 10)

    def test_mul(self):
        tf = Transform(1, 2, 3, 4, 5, 6)

        assert tf * 6 == Transform(6, 12, 18, 24, 30, 36)
        assert 6 * tf == Transform(6, 12, 18, 24, 30, 36)

        with pytest.raises(TypeError):
            _ = tf * "abc"

    def test_matmul(self):
        tf1 = Transform(1, 2, 3, 4, 5, 6)
        tf2 = Transform(6, 5, 4, 3, 2, 1)

        assert tf1 @ tf2 == Transform(a=14, b=11, c=34, d=27, e=56, f=44)
        assert tf2 @ tf1 == Transform(a=21, b=32, c=13, d=20, e=10, f=14)

        with pytest.raises(TypeError):
            _ = tf1 @ 123

    def test_render(self):
        tf = Transform(1, 2, 3, 4, 5, 6)

        assert tf.render(Move(Point(0, 0)))[0] == "1 2 3 4 5 6 cm"

    def test_str(self):
        # don't actually assert the output looks like something in particular, just make
        # sure it doesn't raise, I guess.
        str(Transform(1, 2, 3, 4, 5, 6))


class TestStyleEnums:
    @pytest.mark.parametrize("owner, values, result, guard", parameters.coercive_enums)
    def test_coercive_enums(self, owner, values, result, guard):
        # this is just brute force testing which doesn't seem to be the most intelligent
        # way to do it, but I also don't have a better strategy at all
        for value in values:
            with guard():
                assert owner.coerce(value) is result


class TestStyles:
    @pytest.mark.parametrize("style_name, value", parameters.style_attributes)
    def test_individual_attribute(
        self, auto_pdf_cmp, open_path_drawing, style_name, value
    ):
        setattr(open_path_drawing.style, style_name, value)

        auto_pdf_cmp.draw_path(open_path_drawing)

    def test_stroke_miter_limit(self, auto_pdf_cmp, open_path_drawing):
        open_path_drawing.style.stroke_join_style = "miter"
        open_path_drawing.style.stroke_miter_limit = 1.4

        auto_pdf_cmp.draw_path(open_path_drawing)

    def test_stroke_dash_phase(self, auto_pdf_cmp, open_path_drawing):
        open_path_drawing.style.stroke_dash_pattern = [0.5, 0.5]
        open_path_drawing.style.stroke_dash_phase = 0.5

        auto_pdf_cmp.draw_path(open_path_drawing)

    @pytest.mark.parametrize(
        "blend_mode", parameters.blend_modes, ids=lambda param: f"blend mode {param}"
    )
    def test_blend_modes(
        self, auto_pdf_cmp, open_path_drawing, prepared_blend_path, blend_mode
    ):
        prepared_blend_path.style.blend_mode = blend_mode

        auto_pdf_cmp.draw_path(open_path_drawing)

    def test_serialize_to_pdf_dict(self):
        style = GraphicsStyle()

        style.fill_opacity = 0.5
        style.stroke_opacity = 0.75
        style.blend_mode = "lighten"
        style.stroke_width = 2
        style.stroke_join_style = "round"
        style.stroke_cap_style = "butt"

        assert (
            style.serialize()
            == "<< /Type /ExtGState\n/ca 0.5\n/BM /Lighten\n/CA 0.75\n/LW 2\n/LC 0\n/LJ 1 >>"
        )

    @pytest.mark.parametrize("style_name, value, exception", parameters.invalid_styles)
    def test_bad_style_parameters(self, style_name, value, exception):
        style = GraphicsStyle()
        with pytest.raises(exception):
            setattr(style, style_name, value)

    def test_merge(self):
        style = GraphicsStyle()

        style.fill_opacity = 0.5
        style.stroke_opacity = 0.75
        style.blend_mode = "lighten"
        style.stroke_width = 2
        style.stroke_join_style = "round"
        style.stroke_cap_style = "butt"

        override = GraphicsStyle()

        override.fill_opacity = 0.25
        override.stroke_opacity = 0.1
        override.blend_mode = "hue"

        merged = GraphicsStyle.merge(style, override)

        assert merged.fill_opacity == 0.25
        assert merged.stroke_opacity == 0.1
        assert merged.blend_mode == BlendMode.HUE.value
        assert merged.stroke_width == 2
        assert merged.stroke_join_style == StrokeJoinStyle.ROUND.value
        assert merged.stroke_cap_style == StrokeCapStyle.BUTT.value

    @pytest.mark.parametrize("paint_rule, expected", parameters.paint_rules)
    def test_paint_rule(self, paint_rule, expected):
        style = GraphicsStyle()
        style.paint_rule = paint_rule
        assert style.paint_rule is expected

    def test_paint_rule_resolution(self):
        style = GraphicsStyle()

        style.paint_rule = "auto"

        assert style.resolve_paint_rule() is PathPaintRule.STROKE_FILL_NONZERO

        style.stroke_width = 2
        style.stroke_color = "#000"
        style.fill_color = None
        assert style.resolve_paint_rule() is PathPaintRule.STROKE

        style.fill_color = "#123"
        assert style.resolve_paint_rule() is PathPaintRule.STROKE_FILL_NONZERO

        style.intersection_rule = IntersectionRule.EVENODD
        assert style.resolve_paint_rule() is PathPaintRule.STROKE_FILL_EVENODD

    def test_copy(self):
        style = GraphicsStyle()

        style.fill_opacity = 0.5

        copied = copy.deepcopy(style)

        copied.fill_opacity = 1.0

        assert style is not copied
        assert style.fill_opacity != copied.fill_opacity


def test_path_start_relative():
    """
    Test starting a path with a relative move.
    The parametrized series below can't cover that case.
    """
    point = Point(1, 2)
    start = Move(point)
    element = RelativeLine(Point(3, 4))
    expected = "4 6 l"
    style = GraphicsStyle()
    style.auto_close = False
    rendered, last_item, _ = element.render({}, style, start, point)
    assert rendered == expected
    assert isinstance(last_item, Line)


class TestPathElements:
    @pytest.mark.parametrize(
        "point, element, expected, end_point, end_class", parameters.path_elements
    )
    def test_render(self, point, element, expected, end_point, end_class):
        start = Move(point)
        style = GraphicsStyle()
        style.auto_close = False
        rendered, last_item, _ = element.render({}, style, start, point)

        assert rendered == expected
        assert isinstance(last_item, end_class)
        assert last_item.end_point == pytest.approx(end_point)


class TestDrawingContext:
    def test_add_item(self):
        ctx = DrawingContext()

        ctx.add_item(GraphicsContext())

        with pytest.raises(TypeError):
            ctx.add_item(Move(Point(1, 2)))

    def test_empty_render(self):
        ctx = DrawingContext()
        ctx.add_item(GraphicsContext())

        resource_catalog = ResourceCatalog()
        start = Point(0, 0)
        style = GraphicsStyle()

        result = ctx.render(resource_catalog, start, 1, 10, style)

        assert result == ""

    def test_render(self):
        ctx = DrawingContext()
        ctx.add_item(PaintedPath().line_to(10, 10))

        resource_catalog = ResourceCatalog()
        start = Point(0, 0)
        style = GraphicsStyle()

        result = ctx.render(resource_catalog, start, 1, 10, style)

        assert result == "q 1 0 0 -1 0 10 cm q 0 0 m 10 10 l h B Q Q"

    def test_concurrent_drawing_context(self, auto_pdf):
        with auto_pdf.drawing_context() as _:
            with pytest.raises(FPDFException):
                with auto_pdf.drawing_context() as _:
                    pass


# this is named so it doesn't get picked up by pytest implicitly.
class CommonPathTests:
    path_class = PaintedPath
    comp_index = 0

    @pytest.mark.parametrize(
        "method_calls, elements, rendered", parameters.painted_path_elements
    )
    def test_path_elements(self, method_calls, elements, rendered):
        pth = self.path_class()

        for method, args in method_calls:
            method(pth, *args)

        assert pth._graphics_context.path_items == elements

    def test_implicit_close_insertion(self):
        pth = self.path_class()

        pth.move_to(0, 0)
        pth.line_to(1, 1)
        pth.move_to(2, 2)

        assert pth._closed is True
        assert pth._starter_move == Move(Point(2, 2))
        assert pth._graphics_context.path_items[-1] == ImplicitClose()

    def test_style_property(self):
        pth = self.path_class()
        pth.style.fill_color = "#010203"
        assert pth._graphics_context.style.fill_color == rgb8(1, 2, 3)

    def test_transform_property(self):
        pth = self.path_class()

        tf = Transform.translation(5, 10)
        pth.transform = tf

        assert pth._graphics_context.transform == tf

    def test_transform_group(self):
        pth = self.path_class()
        tf = Transform.translation(1, 2)
        with pth.transform_group(Transform.translation(1, 2)):
            assert pth._root_graphics_context != pth._graphics_context
            assert pth._graphics_context.transform == tf

    def test_auto_close(self):
        pth = self.path_class()
        pth.auto_close = True
        assert pth._root_graphics_context.style.auto_close is True

    def test_paint_rule(self):
        pth = self.path_class()
        pth.paint_rule = "auto"
        assert pth._root_graphics_context.style.paint_rule is PathPaintRule.AUTO

    @pytest.mark.parametrize("rendered", parameters.clipping_path_result)
    def test_clipping_path(self, rendered):
        pth = self.path_class()
        clp = ClippingPath()

        clp.move_to(1, 1).horizontal_line_to(9).vertical_line_to(9).horizontal_line_to(
            1
        ).close()

        pth.clipping_path = clp

        pth.move_to(0, 0)
        pth.line_to(10, 0)
        pth.line_to(5, 10)
        pth.close()

        assert pth._root_graphics_context.clipping_path == clp

        resource_catalog = ResourceCatalog()
        style = GraphicsStyle()
        style.paint_rule = "auto"

        point = Point(0, 0)
        rend, _, __ = pth.render(resource_catalog, style, Move(point), point)
        assert rend == rendered[self.comp_index]

    @pytest.mark.parametrize(
        "method_calls, elements, rendered", parameters.painted_path_elements
    )
    def test_render(self, method_calls, elements, rendered):
        # parameterized this way, this test has a lot of overlap with
        # TestPathElements.test_render, so if one breaks probably the other will. Maybe
        # it's good enough to just check a single render case instead.
        resource_catalog = ResourceCatalog()
        style = GraphicsStyle()
        style.paint_rule = "auto"

        pth = self.path_class()

        for method, args in method_calls:
            method(pth, *args)

        point = Point(0, 0)
        rend, _, __ = pth.render(resource_catalog, style, Move(point), point)
        assert rend == rendered[self.comp_index]

    def test_copy(self):
        # this is a pretty clunky way to test this, but it does make sure that copying
        # works correctly through rendering.
        point = Point(0, 0)
        start = Move(point)
        resource_catalog = ResourceCatalog()
        style = GraphicsStyle()
        style.paint_rule = "auto"

        pth = self.path_class()
        pth.line_to(1, 1)

        pth2 = copy.deepcopy(pth)

        rend1, _, __ = pth.render(resource_catalog, style, start, point)
        rend2, _, __ = pth2.render(resource_catalog, style, start, point)

        pth2.line_to(2, 2)

        rend3, _, __ = pth.render(resource_catalog, style, start, point)
        rend4, _, __ = pth2.render(resource_catalog, style, start, point)

        assert rend1 == rend2 == rend3
        assert rend3 != rend4


class TestPaintedPath(CommonPathTests):
    def test_inheriting_document_properties(self, auto_pdf_cmp):
        auto_pdf_cmp.set_line_width(0.25)
        auto_pdf_cmp.set_dash_pattern(dash=0.25, gap=0.5, phase=0.2)
        auto_pdf_cmp.set_draw_color(255, 0, 0)
        auto_pdf_cmp.set_fill_color(0, 0, 255)

        auto_pdf_cmp.compress = False

        auto_pdf_cmp.rect(1, 1, 8, 2, style="DF")

        with auto_pdf_cmp.new_path() as path:
            path.style.paint_rule = PathPaintRule.STROKE_FILL_NONZERO

            path.rectangle(1, 4, 8, 2)

        with auto_pdf_cmp.new_path() as path:
            path.style.paint_rule = PathPaintRule.STROKE_FILL_NONZERO
            path.style.stroke_width = 0.25
            path.style.stroke_dash_pattern = (0.25, 0.5)
            path.style.stroke_dash_phase = 0.2

            path.rectangle(1, 7, 8, 2)


# This class inherits all of the tests from the PaintedPath test class, just like it
# inherits all of its functionality.
class TestClippingPath(CommonPathTests):
    path_class = ClippingPath
    comp_index = 1


# these tests are all more or less redundant with the PaintedPath tests in terms of
# functionality, but having them implemented may give more information about where a
# regression occurs, if one does.
class TestGraphicsContext:
    def test_copy(self):
        point = Point(0, 0)
        start = Move(point)
        resource_catalog = ResourceCatalog()
        style = GraphicsStyle()
        style.paint_rule = "auto"

        gfx = GraphicsContext()
        gfx.add_item(Move(Point(1, 2)))

        gfx2 = copy.deepcopy(gfx)

        rend1, _, __ = gfx.render(resource_catalog, style, start, point)
        rend2, _, __ = gfx2.render(resource_catalog, style, start, point)

        gfx2.add_item(Line(Point(3, 4)))

        rend3, _, __ = gfx.render(resource_catalog, style, start, point)
        rend4, _, __ = gfx2.render(resource_catalog, style, start, point)

        assert rend1 == rend2 == rend3
        assert rend3 != rend4

        assert rend1 == "q 1 2 m Q"
        assert rend4 == "q 1 2 m 3 4 l Q"

    def test_transform_property(self):
        gfx = GraphicsContext()

        tf = Transform.translation(5, 10)
        gfx.transform = tf

        assert gfx.transform == tf

    def test_clipping_path_property(self):
        gfx = GraphicsContext()
        clp = ClippingPath()

        clp.move_to(1, 1).horizontal_line_to(9).vertical_line_to(9).horizontal_line_to(
            1
        ).close()

        gfx.clipping_path = clp
        gfx.add_item(Move(Point(1, 2)))
        gfx.add_item(Line(Point(3, 4)))

        assert gfx.clipping_path == clp

        point = Point(0, 0)
        start = Move(point)
        resource_catalog = ResourceCatalog()
        style = GraphicsStyle()
        style.paint_rule = "auto"

        rend, _, __ = gfx.render(resource_catalog, style, start, point)
        assert rend == "q 1 1 m 9 1 l 9 9 l 1 9 l h W n 1 2 m 3 4 l Q"

    def test_empty_render(self):
        point = Point(0, 0)
        start = Move(point)
        resource_catalog = ResourceCatalog()
        style = GraphicsStyle()
        style.paint_rule = "auto"

        tf = Transform.translation(5, 10)
        clp = ClippingPath()

        clp.move_to(1, 1).horizontal_line_to(9).vertical_line_to(9).horizontal_line_to(
            1
        ).close()

        gfx = GraphicsContext()
        gfx.clipping_path = clp
        gfx.transform = tf

        rend, _, __ = gfx.render(resource_catalog, style, start, point)
        assert rend == ""

    def test_add_item(self):
        gfx = GraphicsContext()
        gfx.add_item(Move(Point(10, 0)))

        assert gfx.path_items == [Move(Point(10, 0))]

    def test_merge(self):
        gfx1 = GraphicsContext()
        gfx1.add_item(Move(Point(10, 0)))

        gfx2 = GraphicsContext()
        gfx2.add_item(Line(Point(2, 2)))

        gfx1.merge(gfx2)

        assert gfx1.path_items == [
            Move(Point(10, 0)),
            Line(Point(2, 2)),
        ]

    def test_build_render_list(self):
        point = Point(0, 0)
        start = Move(point)
        resource_catalog = ResourceCatalog()
        style = GraphicsStyle()
        style.paint_rule = "auto"

        gfx = GraphicsContext()
        gfx.add_item(Move(Point(1, 2)))
        gfx.add_item(Line(Point(3, 4)))

        rend_list1, _, __ = gfx.build_render_list(resource_catalog, style, start, point)

        assert rend_list1 == ["q", "1 2 m", "3 4 l", "Q"]

        rend_list2, _, __ = gfx.build_render_list(
            resource_catalog, style, start, point, _push_stack=False
        )

        assert rend_list2 == ["1 2 m", "3 4 l"]

    def test_render(self):
        point = Point(0, 0)
        start = Move(point)
        resource_catalog = ResourceCatalog()
        style = GraphicsStyle()
        style.paint_rule = "auto"

        gfx = GraphicsContext()
        gfx.add_item(Move(Point(1, 2)))
        gfx.add_item(Line(Point(3, 4)))

        rend1, _, __ = gfx.render(resource_catalog, style, start, point)

        assert rend1 == "q 1 2 m 3 4 l Q"

        rend2, _, __ = gfx.render(
            resource_catalog, style, start, point, _push_stack=False
        )

        assert rend2 == "1 2 m 3 4 l"


def test_check_page():
    pdf = FPDF(unit="pt")

    with pytest.raises(FPDFException) as no_page_err:
        with pdf.new_path() as _:
            pass

    expected_error_msg = "No page open, you need to call add_page() first"
    assert expected_error_msg == str(no_page_err.value)


def test_blending_images(tmp_path):
    pdf = FPDF(format=(112, 1112))
    pdf.set_margin(0)
    pdf.set_font("Helvetica", size=24)
    pdf.add_page()
    for i, blend_mode in enumerate(BlendMode):
        pdf.image(HERE / "../image/png_test_suite/f04n2c08.png", y=70 * i, h=64)
        with pdf.local_context(blend_mode=blend_mode):
            pdf.image(HERE / "../image/png_test_suite/pp0n6a08.png", y=70 * i, h=64)
        pdf.text(x=0, y=70 * i + 10, text=str(blend_mode))
    assert_pdf_equal(pdf, HERE / "generated_pdf/blending_images.pdf", tmp_path)


def test_invalid_blend_mode():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(ValueError):
        with pdf.local_context(blend_mode="INVALID_VALUE"):
            pass



# pyright: reportUnknownArgumentType=false, reportUnknownMemberType=false
# pylint: disable=fixme,protected-access
"""
This module is in work-in-progress state.
Hint tables / hint streams have not been implemented yet,
and there are a few "TODO" comment remaining.
cf. https://github.com/py-pdf/fpdf2/issues/62
"""

from typing import TYPE_CHECKING, Optional

from .output import ContentWithoutID, OutputProducer, PDFCatalog, PDFHeader, PDFInfo
from .sign import sign_content
from .syntax import (
    DestinationXYZ,
    PDFArray,
    PDFContentStream,
    PDFObject,
    iobj_ref as pdf_ref,
)
from .util import buffer_subst

try:
    from endesive import signer
except ImportError:
    signer = None

if TYPE_CHECKING:
    from .encryption import StandardSecurityHandler

HINT_STREAM_OFFSET_LENGTH_PLACEHOLDER = "0%1%2%3%4%5%6%7%8%9%a%b%c%d"
FIRST_PAGE_END_OFFSET_PLACEHOLDER = "1%2%3%4%5%6%"
MAIN_XREF_1ST_ENTRY_OFFSET_PLACEHOLDER = "2%3%4%5%6%7%"
FILE_LENGTH_PLACEHOLDER = "3%4%5%6%7%8%"


class PDFLinearization(PDFObject):
    def __init__(self, pages_count: int) -> None:
        super().__init__()
        self.linearized = "1"  # Version
        self.n = pages_count
        # Primary hint stream offset and length (part 5):
        self.h = HINT_STREAM_OFFSET_LENGTH_PLACEHOLDER
        self.o = None  # Object number of first page’s page object (part 6)
        self.e = FIRST_PAGE_END_OFFSET_PLACEHOLDER  # Offset of end of first page
        # Offset of first entry in main cross-reference table (part 11):
        self.t = MAIN_XREF_1ST_ENTRY_OFFSET_PLACEHOLDER
        self.l = FILE_LENGTH_PLACEHOLDER  # The length of the entire file in bytes


class PDFXrefAndTrailer(ContentWithoutID):
    PREV_MAIN_XREF_START_PLACEHOLDER = "0%1*2+3-2/1^"

    def __init__(self, output_builder: OutputProducer) -> None:
        self.output_builder = output_builder
        self.count = output_builder.obj_id + 1
        self.start_obj_id = 1
        # Must be set before the call to serialize():
        self.catalog_obj: Optional[PDFCatalog] = None
        self.info_obj: Optional[PDFInfo] = None
        self.first_xref: Optional["PDFXrefAndTrailer"] = None
        self.main_xref: Optional["PDFXrefAndTrailer"] = None
        # Computed at serialize() time based on output_builder.buffer size:
        self.startxref: Optional[str] = None

    @property
    def is_first_xref(self) -> bool:
        return bool(self.main_xref)

    @property
    def is_main_xref(self) -> bool:
        return bool(self.first_xref)

    def serialize(
        self, _security_handler: Optional["StandardSecurityHandler"] = None
    ) -> str:
        builder = self.output_builder
        assert builder is not None
        assert builder.buffer is not None
        out: list[str] = []
        self.startxref = str(len(builder.buffer))
        if self.is_main_xref:
            builder.buffer = buffer_subst(
                builder.buffer,
                self.PREV_MAIN_XREF_START_PLACEHOLDER,
                self.startxref.rjust(12, " "),
            )
        out.append("xref")
        out.append(f"{0 if self.start_obj_id == 1 else self.start_obj_id} {self.count}")
        if not self.is_first_xref:
            out.append("0000000000 65535 f ")
        assert (
            len(builder.offsets) > 1
        ), "TODO: how to know the offsets in the 1st xref at this stage?"
        for obj_id in range(self.start_obj_id, self.start_obj_id + self.count):
            out.append(f"{builder.offsets[obj_id]:010} 00000 n ")
        out.append("trailer")
        out.append("<<")
        if self.is_main_xref:
            assert self.first_xref is not None
            out.append(f"/Size {self.count - self.first_xref.count}")
        else:
            if self.is_first_xref:
                assert self.main_xref is not None
                out.append(f"/Size {self.main_xref.count}")
                out.append(f"/Prev {self.PREV_MAIN_XREF_START_PLACEHOLDER}")
            else:
                out.append(f"/Size {self.count}")
            assert self.catalog_obj is not None
            assert self.info_obj is not None
            out.append(f"/Root {pdf_ref(self.catalog_obj.id)}")
            out.append(f"/Info {pdf_ref(self.info_obj.id)}")
            fpdf = builder.fpdf
            file_id = fpdf.file_id()
            if file_id == -1:
                file_id = fpdf._default_file_id(  # pyright: ignore[reportPrivateUsage]
                    builder.buffer
                )
            if file_id:
                out.append(f"/ID [{file_id}]")
        out.append(">>")
        out.append("startxref")
        startxref = self.startxref
        if self.is_main_xref:
            assert self.first_xref is not None
            startxref = str(self.first_xref.startxref or "0")
        if self.is_first_xref:
            startxref = "0"
        out.append(startxref)
        out.append("%%EOF")
        return "\n".join(out)


class PDFHintStream(PDFContentStream):
    def __init__(self, contents: bytes, compress: bool = False) -> None:
        super().__init__(contents=contents, compress=compress)
        self.s = None  # (Required) Shared object hint table
        self.t = None  # (Present only if thumbnail images exist) Thumbnail hint table
        self.o = None  # (Present only if a document outline exists) Outline hint table
        self.a = None  # (Present only if article threads exist) Thread information hint table
        self.e = None  # (Present only if named destinations exist) Named destination hint table
        self.v = None  # (Present only if an interactive form dictionary exists) Interactive form hint table
        self.i = None  # (Present only if a document information dictionary exists) Information dictionary hint table
        self.c = None  # (Present only if a logical structure hierarchy exists; PDF 1.3) Logical structure hint table
        self.l = None  # (PDF 1.3) Page label hint table
        self.r = None  # (Present only if a renditions name tree exists; PDF 1.5) Renditions name tree hint table
        self.b = None  # (Present only if embedded file streams exist; PDF 1.5) Embedded file stream hint table


class LinearizedOutputProducer(OutputProducer):
    def bufferize(self) -> bytearray:
        fpdf = self.fpdf

        # 1. Setup - Insert all PDF objects
        #    (in the order required to build a linearized PDF),
        #    and assign unique consecutive numeric IDs to all of them

        # Part 1: Header
        self.pdf_objs.append(PDFHeader(fpdf.pdf_version))
        # Part 2: Linearization parameter dictionary
        linearization_obj = PDFLinearization(fpdf.pages_count)
        self._add_pdf_obj(linearization_obj)
        # Part 3: First-page cross-reference table and trailer
        first_xref = PDFXrefAndTrailer(self)
        self.pdf_objs.append(first_xref)
        # Part 4: Document catalogue and other required document-level objects
        catalog_obj = self._add_catalog()
        # Part 5: Primary hint stream (may precede or follow part 6)
        hint_stream_obj = PDFHintStream("".encode("latin-1"))  # TODO
        self.pdf_objs.append(hint_stream_obj)
        # Part 6: First-page section (may precede or follow part 5)
        page_objs = self._add_pages(slice(0, 1))
        # The following objects shall be contained in the first-page section:
        #   + This page object shall explicitly specify all required attributes, e.g. Resources, MediaBox
        #   + The entire outline hierarchy, if the PageMode entry in the catalogue is UseOutlines
        #   + All objects that the page object refers to [including] Contents, Resources, Annots
        # TODO

        first_xref.count = self.obj_id + 1
        first_xref_pdf_objs = list(self.pdf_objs)
        self.obj_id = 0

        # Part 7: Remaining pages
        page_objs.extend(self._add_pages(slice(1, None)))
        # Part 8: Shared objects for all pages except the first
        # = resources, that are referenced from more than one page but [not] from the first page
        pages_root_obj = self._add_pages_root()
        sig_annotation_obj = self._add_annotations_as_objects()
        img_objs_per_index = self._add_images()
        gfxstate_objs_per_name = self._add_gfxstates()
        shading_objs_per_name = self._add_shadings()
        pattern_objs_per_name = self._add_patterns()
        font_objs_per_index = self._add_fonts(
            img_objs_per_index,
            gfxstate_objs_per_name,
            pattern_objs_per_name,
        )
        resources_dict_obj = self._add_resources_dict(
            font_objs_per_index,
            img_objs_per_index,
            gfxstate_objs_per_name,
            shading_objs_per_name,
            pattern_objs_per_name,
        )
        # Part 9: Objects not associated with pages, if any
        for embedded_file in fpdf.embedded_files:
            self._add_pdf_obj(embedded_file, "embedded_files")
        struct_tree_root_obj = self._add_structure_tree()
        outline_dict_obj, outline_items = self._add_document_outline()
        xmp_metadata_obj = self._add_xmp_metadata()
        info_obj = self._add_info()
        # Part 11: Main cross-reference table and trailer
        main_xref = PDFXrefAndTrailer(self)
        self.pdf_objs.append(main_xref)

        # Re-assigning IDs of all PDF objects in the 1st xref table:
        first_xref.start_obj_id = self.obj_id + 1
        for pdf_obj in first_xref_pdf_objs:
            if (
                not isinstance(pdf_obj, ContentWithoutID)
                and pdf_obj is not hint_stream_obj
            ):
                self.obj_id += 1
                pdf_obj.id = self.obj_id
        # The hint streams shall be assigned the last object numbers in the file:
        self.obj_id += 1
        hint_stream_obj.id = self.obj_id

        # 2. Plumbing - Inject all PDF object references required:
        linearization_obj.o = page_objs[0].id  # type: ignore[assignment]
        pages_root_obj.kids = PDFArray(page_objs)
        self._finalize_catalog(
            catalog_obj,
            pages_root_obj=pages_root_obj,
            first_page_obj=page_objs[0],
            sig_annotation_obj=sig_annotation_obj,
            xmp_metadata_obj=xmp_metadata_obj,
            struct_tree_root_obj=struct_tree_root_obj,
            outline_dict_obj=outline_dict_obj,
        )
        dests: list[DestinationXYZ] = []
        for page_obj in page_objs:
            page_obj.parent = pages_root_obj
            page_obj.resources = resources_dict_obj
            assert page_obj.annots is not None
            for annot in page_obj.annots:
                if annot.dest:
                    dests.append(annot.dest)
                if annot.a and hasattr(annot.a, "dest"):
                    dests.append(annot.a.dest)
            if not page_obj.annots:
                # Avoid serializing an empty PDFArray:
                page_obj.annots = None
        for outline_item in outline_items:
            if outline_item.dest is not None:
                dests.append(outline_item.dest)
        # Assigning the .page_ref property of all Destination objects:
        for dest in dests:
            dest.page_ref = pdf_ref(  # pyright: ignore[reportAttributeAccessIssue]
                page_objs[
                    dest.page_number - 1  # pyright: ignore[reportAttributeAccessIssue]
                ].id
            )
        for struct_elem in fpdf.struct_builder.doc_struct_elem.k:
            struct_elem.pg = page_objs[struct_elem.page_number() - 1]
        main_xref.first_xref = first_xref
        first_xref.main_xref = main_xref
        for xref in [main_xref, first_xref]:
            xref.catalog_obj = catalog_obj
            xref.info_obj = info_obj

        # 3. Serializing - Append all PDF objects to the buffer:
        assert (
            not self.buffer
        ), f"Nothing should have been appended to the .buffer at this stage: {self.buffer}"
        assert (
            not self.offsets
        ), f"No offset should have been set at this stage: {len(self.offsets)}"
        for pdf_obj in self.pdf_objs:
            if isinstance(pdf_obj, ContentWithoutID):
                # top header, xref table & trailer:
                trace_label = None
            else:
                self.offsets[pdf_obj.id] = len(self.buffer)
                trace_label = self.trace_labels_per_obj_id.get(pdf_obj.id)
            if trace_label:
                with self._trace_size(trace_label):
                    self._out(pdf_obj.serialize())
            else:
                self._out(pdf_obj.serialize())
        self._log_final_sections_sizes()

        # Now that the file size & all the offsets are known,
        # substitute the values of the Linearization properties:
        hs1_offset = self.offsets[hint_stream_obj.id]
        hs1_length = len(hint_stream_obj.serialize())
        self.buffer = buffer_subst(
            self.buffer,
            HINT_STREAM_OFFSET_LENGTH_PLACEHOLDER,
            f"[{hs1_offset: 12d} {hs1_length: 12d}]",
        )
        self.buffer = buffer_subst(
            self.buffer,
            FIRST_PAGE_END_OFFSET_PLACEHOLDER,
            f"{self.offsets[page_objs[0].id + 1]: 12d}",
        )
        self.buffer = buffer_subst(
            self.buffer,
            MAIN_XREF_1ST_ENTRY_OFFSET_PLACEHOLDER,
            f"{self.offsets[main_xref.start_obj_id]: 12d}",
        )
        self.buffer = buffer_subst(
            self.buffer,
            FILE_LENGTH_PLACEHOLDER,
            f"{len(self.buffer): 12d}",
        )

        if fpdf._sign_key:  # pyright: ignore[reportPrivateUsage]
            self.buffer = sign_content(
                signer,  # pyright: ignore[reportArgumentType]
                self.buffer,
                fpdf._sign_key,  # pyright: ignore[reportPrivateUsage]
                fpdf._sign_cert,  # pyright: ignore[reportPrivateUsage]
                fpdf._sign_extra_certs,  # pyright: ignore[reportPrivateUsage]
                fpdf._sign_hashalgo,  # pyright: ignore[reportPrivateUsage]
                fpdf._sign_time,  # pyright: ignore[reportPrivateUsage]
            )

        return self.buffer



#!/usr/bin/env python3
# USAGE: ./add_on_page_with_pypdf.py src_file.pdf dest_file.pdf
import io, sys
from contextlib import contextmanager

# Before 2.8.2 use: from fpdf.util import get_scale_factor
from fpdf import FPDF, get_scale_factor
from pypdf import PdfReader, PdfWriter

IN_FILEPATH = sys.argv[1]
OUT_FILEPATH = sys.argv[2]


@contextmanager
def add_to_page(reader_page, unit="mm"):
    k = get_scale_factor(unit)
    format = (reader_page.mediabox[2] / k, reader_page.mediabox[3] / k)
    pdf = FPDF(format=format, unit=unit)
    pdf.add_page()
    yield pdf
    page_overlay = PdfReader(io.BytesIO(pdf.output())).pages[0]
    reader_page.merge_page(page2=page_overlay)


reader = PdfReader(IN_FILEPATH)
with add_to_page(reader.pages[0]) as pdf:
    pdf.set_font("times", style="B", size=30)
    pdf.text(50, 150, "Hello World!")

writer = PdfWriter()
writer.append_pages_from_reader(reader)
writer.write(OUT_FILEPATH)
